# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""

from typing import Dict, Tuple, List

import urllib3
import logging as log
from IPy import IP
from infoblox_client import connector

from infoblox_discovery.infoblox_dhcp import DHCP, dhcp_factory
from infoblox_discovery.infoblox_dns_server import DNSServer, dns_server_factory
from infoblox_discovery.infoblox_zone import Zone, zone_factory
from infoblox_discovery.infoblox_member import Member, member_factory
from infoblox_discovery.infoblox_node import Node, node_factory
from infoblox_discovery.infoblox_webendpoint import WebEndpoint, webendpoint_factory
from infoblox_discovery.exceptions import DiscoveryException

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

MEMBERS = "members"
ZONES = "zones"
DHCP_RANGES = "dhcp_ranges"


class InfoBlox:
    def __init__(self, config):

        self.master = config.get('master')

        self.exclude_ranges = config.get('exclude_ranges')
        self.exclusions: Dict[str, List[str]] = {}
        self.inclusions: Dict[str, List[str]] = {}

        if config.get('exclusion_labels') is not None:
            for exclustion_key, exclustions in config.get('exclusion_labels').items():
                if exclustion_key not in [MEMBERS, ZONES, DHCP_RANGES]:
                    raise DiscoveryException(f"Invalid exclusion label {exclustion_key}")

                self.exclusions[exclustion_key] = exclustions

        if config.get('inclusion_labels') is not None:
            for inclustion_key, inclustions in config.get('inclusion_labels').items():
                if inclustion_key not in [MEMBERS, ZONES, DHCP_RANGES]:
                    raise DiscoveryException(f"Invalid inclusion label {inclustion_key}")

                self.inclusions[inclustion_key] = inclustions

        self.opts = {'host': config.get('master'),
                     'username': config.get('username'),
                     'password': config.get('password'),
                     'wapi_version': config.get('wapi_version'),
                     'http_request_timeout': config.get('timeout', 60)}
        self.conn = connector.Connector(self.opts)

    def get_infoblox_members(self) -> Tuple[Dict[str, Member], Dict[str, Node], Dict[str, DNSServer]]:
        return_fields_member = ['host_name', 'service_status', 'platform', 'enable_ha', 'node_info', 'ntp_setting',
                                'extattrs']
        try:
            members_data = self.conn.get_object('member', return_fields=return_fields_member)
        except Exception as err:
            log.error("Fetch members", extra={"error": str(err)})
            raise DiscoveryException("Could not fetch members")

        members: Dict[str, Member] = {}
        nodes: Dict[str, Node] = {}
        dns_servers: Dict[str: DNSServer] = {}

        for member_data in members_data:
            if self.validate_exclusion(member_data['extattrs'], [MEMBERS]):
                log.info(f"Exclude infoblox member {member_data['host_name']}")
                continue

            member = member_factory(member_data, self.master)
            members[member.host_name] = member

            if member.enable_ha == 'true':
                for node in member_data['node_info']:
                    node = node_factory(node, member.host_name, self.master)
                    nodes[node.ip] = node

            for service in member_data['service_status']:
                if service['service'] == 'DNS' and service['status'] == 'WORKING':
                    dns_server = dns_server_factory(member.host_name, self.master)
                    dns_servers[member.host_name] = dns_server

        log.info("Discovered from object member", extra={"members_infoblox": len(members_data), "members_discovery": len(members), "nodes_discovery": len(nodes), "dns_discovery": len(dns_servers)})
        return members, nodes, dns_servers

    def validate_exclusion(self, extattrs, exclutions: List[str]) -> bool:
        if len(extattrs.keys()) > 0:
            log.debug(f"Extattrs {extattrs}")

        # First check inclusions
        inclusion_exists = False
        for inclusion in exclutions:
            if inclusion in self.inclusions:
                inclusion_exists = True
                try:
                    for inclusion_key in self.inclusions[inclusion]:
                        if inclusion_key in extattrs and extattrs[inclusion_key]['value'] == 'True':
                            return False
                    return True
                except Exception as err:
                    log.error(f"Validate inclusion - {str(err)}")

        # Second check exclusions
        if inclusion_exists:
            # If inclusion exists and not matched, do not execute exclude logic
            return True
        for exclusion in exclutions:
            if exclusion in self.exclusions:
                try:
                    for exclusion_key in self.exclusions[exclusion]:
                        if exclusion_key in extattrs and extattrs[exclusion_key]['value'] == 'True':
                            return True
                    return False
                except Exception as err:
                    log.error(f"Validate exclusion - {str(err)}")

    def get_infoblox_zones(self) -> Dict[str, Zone]:
        return_fields_range = ['fqdn', 'disable', 'extattrs']
        query = {'view': 'External'}
        try:
            zones_data = self.conn.get_object('zone_auth', query, return_fields=return_fields_range)
        except Exception as err:
            log.error(f"Could not fetch zones - {str(err)}")
            raise DiscoveryException("Could not fetch zones")

        all_zones: Dict[str: Zone] = {}
        disabled_zones = 0
        for zone_data in zones_data:
            log.debug(f"Dns zone {zone_data['fqdn']}")
            if self.validate_exclusion(zone_data['extattrs'], [ZONES]) or \
                    'disable' in zone_data and zone_data['disable']:
                if 'disable' in zone_data and zone_data['disable']:
                    disabled_zones += 1
                log.debug(f"Exclude dns zone {zone_data['fqdn']}")
                continue

            zone = {}
            if '/' in zone_data['fqdn']:
                ip = IP(zone_data['fqdn'])
                log.debug(f"dns zone {ip.reverseName()}")

                zone['name'] = ip.reverseName()
                zone['address'] = ip.reverseName()
            else:
                log.debug(f"dns zone {zone_data['fqdn']} - {zone_data['fqdn'].encode('idna').decode('utf-8')}")
                zone['name'] = zone_data['fqdn']
                zone['address'] = zone_data['fqdn'].encode('idna').decode("utf-8")

            z = zone_factory(zone['name'], self.master)
            all_zones[z.zone] = z
        log.info("Discovered from object zone_auth", extra={"zones_infoblox": len(zones_data), "zone_disabled": disabled_zones, "zones_discovery": len(all_zones)})
        return all_zones

    def get_infoblox_dhcp_ranges(self) -> Dict[str, DHCP]:
        return_fields_range = ['network', 'dhcp_utilization', 'dhcp_utilization_status', 'extattrs']
        query = {'network_view': 'default'}

        try:
            dhcp_ranges_data = self.conn.get_object('range', query, return_fields=return_fields_range, paging=True)
        except Exception as err:
            log.error(f"Could not get dhcp ranges, {str(err)}")
            raise DiscoveryException("Could not fetch dhcp ranges")
        dhcp_ranges: Dict[str, DHCP] = {}
        for dhcp_range in dhcp_ranges_data:
            if self.validate_exclusion(dhcp_range['extattrs'], [DHCP_RANGES]):
                log.debug(f"Exclude dhcp scope {dhcp_range['network']}")
                continue

            # Remove all configured scopes
            range = int(dhcp_range['network'].split('/')[1])
            if range in self.exclude_ranges:
                continue

            dhcp = dhcp_factory(dhcp_range['network'], self.master)
            dhcp_ranges[dhcp.network] = dhcp

        log.info("Discovered from object range", extra={"dhcp_ranges_infoblox": len(dhcp_ranges_data), "dhcp_ranges_discovery": len(dhcp_ranges)})
        return dhcp_ranges

    def get_web_endpoints_by_networks(self, network) -> Dict[str, WebEndpoint]:

        fqdn_by_network = self._get_fqdn_by_network(network)
        web_endpoints: Dict[str, WebEndpoint] = {}
        for fqdn in fqdn_by_network:
            res = self._get_endpoint(fqdn)
            for dns in res:
                if 'External' in dns['_ref'] and 'dns_aliases' in dns:
                    for alias in dns['dns_aliases']:
                        web_endpoints[alias] = webendpoint_factory(alias, master=self.master)
        log.info("Discovered from object record:host", extra={"web_endpoints_discovery": len(web_endpoints)})
        return web_endpoints

    def _get_fqdn_by_network(self, network):
        return_fields_range = ['ip_address,names', 'objects', 'types']
        query = {'network': network}
        all_names = self.conn.get_object('ipv4address', query, return_fields=return_fields_range)

        names = []
        for name in all_names:
            if 'HOST' in name['types']:
                names.extend(name['names'])

        log.info("Discovered from object ipv4address", extra={"fqdns_discovery": len(names)})
        return names

    def _get_endpoint(self, dns_fqdn):
        return_fields_range = ['dns_aliases']
        query = {'name': dns_fqdn}
        dns = self.conn.get_object('record:host', query, return_fields=return_fields_range)
        return dns



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""


class DiscoveryException(Exception):
    def __init__(self, message: str = "", status: int = 503, exp: Exception = None):
        self.message = message
        self.status = status
        self.exp = exp



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""
import datetime
import json
import logging
import math
import os
import secrets
import time
from typing import List, Any

import uvicorn
from uvicorn.config import LOGGING_CONFIG
import yaml
from apscheduler.schedulers.background import BackgroundScheduler
from fastapi import FastAPI, Request, Response, Depends, HTTPException, status
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from prometheus_client import CollectorRegistry, Gauge
from prometheus_client.exposition import CONTENT_TYPE_LATEST
from prometheus_client.utils import INF, MINUS_INF
from prometheus_fastapi_instrumentator import Instrumentator

from infoblox_discovery.api import InfoBlox
from infoblox_discovery.cache import Cache, MEMBERS, NODES, ZONES, DHCP_RANGES, DNS_SERVERS, MASTER, WEB_ENDPOINTS, VALID_TYPES
from infoblox_discovery.collector import InfobloxCollector
from infoblox_discovery.environments import DISCOVERY_BASIC_AUTH_USERNAME, DISCOVERY_BASIC_AUTH_PASSWORD, \
    DISCOVERY_BASIC_AUTH_ENABLED, DISCOVERY_HOST, DISCOVERY_PORT, DISCOVERY_FETCH_INTERVAL
from infoblox_discovery.environments import DISCOVERY_CONFIG
from infoblox_discovery.exceptions import DiscoveryException
import logging as log


MIME_TYPE_TEXT_HTML = 'text/html'
MIME_TYPE_APPLICATION_JSON = 'application/json'

app = FastAPI()


def fill_cache():
    """
    Collect data from Infoblox
    The configuration file is read every time
    :return:
    """

    with open(os.getenv(DISCOVERY_CONFIG, 'config.yml'), 'r') as config_file:
        try:
            # Converts yaml document to python object
            config = yaml.safe_load(config_file)

        except yaml.YAMLError as err:
            log.error("Can not open configuration file", extra={"file_name": config_file.name, "error": str(err)})
            return

    cache = Cache()
    for ib in config.get('infoblox'):
        start_time = time.time()
        try:
            infoblox = InfoBlox(ib)
        except DiscoveryException as err:
            log.error("Failed to create infoblox connection", extra={"error": str(err), "master": ib.get(MASTER, 'n/a')})
            continue
        try:
            if MEMBERS in ib.get('discovery'):
                try:
                    members, nodes, dns_servers = infoblox.get_infoblox_members()
                    cache.put(ib[MASTER], MEMBERS, list(members.values()))
                    cache.put(ib[MASTER], NODES, list(nodes.values()))
                    cache.put(ib[MASTER], DNS_SERVERS, list(dns_servers.values()))
                except DiscoveryException:
                    log.error("Failed to get members", extra={"master": ib[MASTER]})

            if ZONES in ib.get('discovery'):
                try:
                    dns = infoblox.get_infoblox_zones()
                    cache.put(ib[MASTER], ZONES, list(dns.values()))
                except DiscoveryException:
                    log.error("Failed to get zones", extra={"master": ib[MASTER]})

            if DHCP_RANGES in ib.get('discovery'):
                try:
                    dhcp_ranges = infoblox.get_infoblox_dhcp_ranges()
                    cache.put(ib[MASTER], DHCP_RANGES, list(dhcp_ranges.values()))
                except DiscoveryException:
                    log.error("Failed to get dhcp ranges", extra={"master": ib[MASTER]})

            if WEB_ENDPOINTS in ib.get('discovery') and ib.get(WEB_ENDPOINTS):
                try:
                    for network in ib.get(WEB_ENDPOINTS).get('networks'):
                        web_endpoints = infoblox.get_web_endpoints_by_networks(network)
                        cache.put(ib[MASTER], WEB_ENDPOINTS, list(web_endpoints.values()))
                except DiscoveryException:
                    log.error("Failed to get dhcp ranges", extra={"master": ib[MASTER]})

            end_time = time.time()
            cache.set_collect_time(ib[MASTER], int(end_time - start_time))
            log.info("Collect infoblox discovery", extra={"master": ib[MASTER], "exec_time_seconds": end_time - start_time})
        except DiscoveryException:
            cache.inc_collect_count_failed(ib[MASTER])
        finally:
            cache.inc_collect_count(ib[MASTER])


@app.on_event("startup")
async def run_scheduler():
    sch = BackgroundScheduler()
    sch.start()
    sch.add_job(fill_cache, 'date', run_date=datetime.datetime.now())
    sch.add_job(fill_cache, 'interval', seconds=int(os.getenv(DISCOVERY_FETCH_INTERVAL, '3600')))

Instrumentator().instrument(app).expose(app=app, endpoint="/exporter-metrics")

security = HTTPBasic()


async def optional_security(request: Request):
    if os.getenv(DISCOVERY_BASIC_AUTH_ENABLED) and os.getenv(DISCOVERY_BASIC_AUTH_ENABLED) == "true":
        return await security(request)
    else:
        return None


async def basic_auth(credentials: HTTPBasicCredentials = Depends(optional_security)) -> bool:

    if not os.getenv(DISCOVERY_BASIC_AUTH_USERNAME) or not os.getenv(DISCOVERY_BASIC_AUTH_USERNAME):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect credentials",
            headers={"WWW-Authenticate": "Basic"},
        )

    current_username_bytes = credentials.username.encode("utf8")
    correct_username_bytes = bytes(os.getenv(DISCOVERY_BASIC_AUTH_USERNAME), 'utf-8')
    is_correct_username = secrets.compare_digest(
        current_username_bytes, correct_username_bytes
    )
    current_password_bytes = credentials.password.encode("utf8")
    correct_password_bytes = bytes(os.getenv(DISCOVERY_BASIC_AUTH_PASSWORD), 'utf-8')
    is_correct_password = secrets.compare_digest(
        current_password_bytes, correct_password_bytes
    )
    if not (is_correct_username and is_correct_password):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect credentials",
            headers={"WWW-Authenticate": "Basic"},
        )
    return True


def floatToGoString(d):
    d = float(d)
    if d == INF:
        return '+Inf'
    elif d == MINUS_INF:
        return '-Inf'
    elif math.isnan(d):
        return 'NaN'
    else:
        s = repr(d)
        dot = s.find('.')
        # Go switches to exponents sooner than Python.
        # We only need to care about positive values for le/quantile.
        if d > 0 and dot > 6:
            mantissa = '{0}.{1}{2}'.format(s[0], s[1:dot], s[dot + 1:]).rstrip('0.')
            return '{0}e+0{1}'.format(mantissa, dot - 1)
        return s


def generate_latest(metrics_list: list):
    """
    Returns the metrics from the registry in text format as a string
    :param metrics_list:
    :return:
    """
    """"""

    def sample_line(line):
        if line.labels:
            labelstr = '{{{0}}}'.format(','.join(
                ['{0}="{1}"'.format(
                    k, v.replace('\\', r'\\').replace('\n', r'\n').replace('"', r'\"'))
                    for k, v in sorted(line.labels.items())]))
        else:
            labelstr = ''
        timestamp = ''
        if line.timestamp is not None:
            # Convert to milliseconds.
            timestamp = ' {0:d}'.format(int(float(line.timestamp) * 1000))
        return '{0}{1} {2}{3}\n'.format(
            line.name, labelstr, floatToGoString(line.value), timestamp)

    output = []
    for metric in metrics_list:
        try:
            mname = metric.name
            mtype = metric.type
            # Munging from OpenMetrics into Prometheus format.
            if mtype == 'counter':
                mname = mname + '_total'
            elif mtype == 'info':
                mname = mname + '_info'
                mtype = 'gauge'
            elif mtype == 'stateset':
                mtype = 'gauge'
            elif mtype == 'gaugehistogram':
                # A gauge histogram is really a gauge,
                # but this captures the structure better.
                mtype = 'histogram'
            elif mtype == 'unknown':
                mtype = 'untyped'

            output.append('# HELP {0} {1}\n'.format(
                mname, metric.documentation.replace('\\', r'\\').replace('\n', r'\n')))
            output.append('# TYPE {0} {1}\n'.format(mname, mtype))

            om_samples = {}
            for s in metric.samples:
                for suffix in ['_created', '_gsum', '_gcount']:
                    if s.name == metric.name + suffix:
                        # OpenMetrics specific sample, put in a gauge at the end.
                        om_samples.setdefault(suffix, []).append(sample_line(s))
                        break
                else:
                    output.append(sample_line(s))
        except Exception as exception:
            exception.args = (exception.args or ('',)) + (metric,)
            raise

        for suffix, lines in sorted(om_samples.items()):
            output.append('# HELP {0}{1} {2}\n'.format(metric.name, suffix,
                                                       metric.documentation.replace('\\', r'\\').replace('\n', r'\n')))
            output.append('# TYPE {0}{1} gauge\n'.format(metric.name, suffix))
            output.extend(lines)
    return ''.join(output).encode('utf-8')


@app.get('/alive')
async def alive(request: Request):
    return Response("infoblox_discovery alive!", status_code=status.HTTP_200_OK, media_type=MIME_TYPE_TEXT_HTML)


@app.get('/metrics')
async def get_metrics(auth: str = Depends(basic_auth)):
    start_time = time.time()
    cache = Cache()
    registry = CollectorRegistry()

    try:
        infoblox_collector = InfobloxCollector(cache)

        registry.register(infoblox_collector)

        duration = Gauge('infoblox_scrape_duration_seconds', 'Time spent processing request', registry=registry)

        duration.set(time.time() - start_time)

        infoblox_metrics = generate_latest(await infoblox_collector.collect())

        duration.set(time.time() - start_time)
        return Response(infoblox_metrics, status_code=200, media_type=CONTENT_TYPE_LATEST)
    except DiscoveryException as err:
        log.error("Failed to get metrics", extra={"error": str(err)})
        return Response(err.message, status_code=err.status, media_type=MIME_TYPE_TEXT_HTML)
    except Exception as err:
        log.error("Failed to get metrics", extra={"error": str(err)})
        return Response("Internal server error for - please check logs", status_code=500,
                        media_type=MIME_TYPE_TEXT_HTML)


@app.get('/prometheus-sd-targets')
async def discovery(master: str, type: str, auth: str = Depends(basic_auth)):
    try:
        if type not in VALID_TYPES:
            return Response(json.dumps({'error': 'Not a valid type', 'valid_types': VALID_TYPES}, indent=4), status_code=status.HTTP_400_BAD_REQUEST, media_type=MIME_TYPE_APPLICATION_JSON)
        cache = Cache()
        data = cache.get(master, type)

        prometheus_sd: List[Any] = []
        for d in data:
            prometheus_sd.append(d.as_prometheus_file_sd_entry())

        targets = json.dumps(prometheus_sd, indent=4)
        return Response(targets, status_code=status.HTTP_200_OK, media_type=MIME_TYPE_APPLICATION_JSON)
    except Exception as err:
        log.error("Failed to get prometheus sd targets", extra={"error": str(err)})
        return Response(None, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, media_type=MIME_TYPE_APPLICATION_JSON)


def http_service_discovery():
    logging.Formatter.converter = time.gmtime
    log_config = LOGGING_CONFIG.copy()
    log_config["formatters"]["default"]["fmt"] = "at=%(levelname)s when=%(asctime)s msg=\"%(message)s\""
    log_config["formatters"]["default"]["datefmt"] = "%Y-%m-%dT%H:%M:%SZ"
    log_config["formatters"]["access"]["fmt"] = "at=%(levelname)s when=%(asctime)s msg=\"%(client_addr)s - %(request_line)s\" status=%(status_code)s"
    log_config["formatters"]["access"]["datefmt"] = "%Y-%m-%dT%H:%M:%SZ"

    uvicorn.run(app, host=os.getenv(DISCOVERY_HOST, "0.0.0.0"), port=os.getenv(DISCOVERY_PORT, 9694),
                log_config=log_config)



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""

import os
import time
import logging as log
from typing import Dict, List, Any
from infoblox_discovery.environments import DISCOVERY_CACHE_TTL


MEMBERS = 'members'
NODES = 'nodes'
ZONES = 'zones'
DNS_SERVERS = 'dns_servers'
DHCP_RANGES = 'dhcp_ranges'
WEB_ENDPOINTS = 'web_endpoints'
VALID_TYPES = [MEMBERS, NODES, ZONES, DHCP_RANGES, DNS_SERVERS, WEB_ENDPOINTS]

MASTER = 'master'


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Cache(metaclass=Singleton):

    def __init__(self):
        self._ttl: int = int(os.getenv(DISCOVERY_CACHE_TTL, "7200"))
        self._expire: int = 0
        # master->type-> data
        self._collect_count: Dict[str, int] = {}
        self._collect_time: Dict[str, int] = {}
        self._collect_count_failed: Dict[str, int] = {}
        self._cache: Dict[str, Dict[str, List]] = {}

    def put(self, master: str, type: str, data: List[Any]):
        self._expire = time.time() + self._ttl
        if master not in self._cache:
            self._cache[master] = {}
        self._cache[master][type] = data

    def get(self, master: str, type: str) -> List[Any]:
        if time.time() < self._expire and master in self._cache and type in self._cache[master]:
            log.info("Cache", extra={"hit": True})
            return self._cache[master][type]
        log.info("Cache", extra={"hit": False})
        return []

    def get_all(self) -> Dict[str, Dict[str, List]]:
        return self._cache

    def set_collect_time(self, master, collect_time: int):
        self._collect_time[master] = collect_time

    def get_collect_time(self) -> Dict[str, int]:
        return self._collect_time

    def inc_collect_count(self, master):
        if master not in self._collect_count:
            self._collect_count[master] = 0
        self._collect_count[master] += 1

    def get_collect_count(self) -> Dict[str, int]:
        return self._collect_count

    def inc_collect_count_failed(self, master):
        if master not in self._collect_count_failed:
            self._collect_count_failed[master] = 0
        self._collect_count_failed[master] += 1

    def get_collect_count_failed(self) -> Dict[str, int]:
        return self._collect_count_failed



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""

from typing import List

from prometheus_client.registry import Collector, Metric

from infoblox_discovery.cache import Cache
from infoblox_discovery.metrics import InfobloxMetrics


def to_list(metric_generator) -> List[Metric]:
    metrics = []
    for metric in metric_generator:
        if metric.samples:
            # Only append if the metric has a list of Samples
            metrics.append(metric)
    return metrics


class InfobloxCollector(Collector):
    def __init__(self, cache: Cache):
        self.cache = cache

    async def collect(self):
        all_module_metrics = []

        transformer = InfobloxMetrics(self.cache)
        transformer.parse()
        t = to_list(transformer.metrics())
        all_module_metrics.extend(t)

        return all_module_metrics



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""

import argparse
import time
import logging
from logfmter import Logfmter
from infoblox_discovery.file_service_discovery import file_service_discovery
from infoblox_discovery.http_service_discovery import http_service_discovery

logging.Formatter.converter = time.gmtime
formatter = Logfmter(
    keys=["at", "when"],
    mapping={"at": "levelname", "when": "asctime"},
    datefmt="%Y-%m-%dT%H:%M:%SZ",
)
handler = logging.StreamHandler()
handler.setFormatter(formatter)

logging.basicConfig(handlers=[handler], level=logging.INFO)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Infoblox service discovery')
    parser.add_argument('--server', action='store_true',
                        help='Start in http service discovery mode',
                        dest='server')
    args = vars(parser.parse_args())

    if args['server']:
        http_service_discovery()
    else:
        file_service_discovery()



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""

from typing import Dict, List

from prometheus_client.core import GaugeMetricFamily
from prometheus_client.metrics_core import Metric, CounterMetricFamily

from infoblox_discovery.cache import Cache, MASTER, MEMBERS, NODES, ZONES, DHCP_RANGES


from infoblox_discovery.transform import Transform, LabelsBase


class IBMetricDefinition:
    prefix = 'infoblox_'
    help_prefix = 'Infoblox '

    class Labels(LabelsBase):
        def __init__(self):
            super().__init__()
            self.labels = {MASTER: ""}

    @staticmethod
    def metrics_definition() -> Dict[str, Metric]:
        common_labels = IBMetricDefinition.Labels().get_label_keys()

        metric_definition = {
            "cache_collect":
                CounterMetricFamily(name=f"{IBMetricDefinition.prefix}cache_collect",
                                    documentation=f"{IBMetricDefinition.help_prefix}total collect count",
                                    labels=common_labels),
            "cache_collect_failed":
                CounterMetricFamily(name=f"{IBMetricDefinition.prefix}cache_collect_failed",
                                    documentation=f"{IBMetricDefinition.help_prefix}total failed collect count",
                                    labels=common_labels),
            "cache_collect_time":
                GaugeMetricFamily(name=f"{IBMetricDefinition.prefix}cache_collect_time",
                                  documentation=f"{IBMetricDefinition.help_prefix}time to collect",
                                  labels=common_labels),
            "cache_members":
                CounterMetricFamily(name=f"{IBMetricDefinition.prefix}cache_members",
                                    documentation=f"{IBMetricDefinition.help_prefix}number of members",
                                    labels=common_labels),
            "cache_nodes":
                CounterMetricFamily(name=f"{IBMetricDefinition.prefix}cache_nodes",
                                    documentation=f"{IBMetricDefinition.help_prefix}number of nodes",
                                    labels=common_labels),
            "cache_zones":
                CounterMetricFamily(name=f"{IBMetricDefinition.prefix}cache_zones",
                                    documentation=f"{IBMetricDefinition.help_prefix}number of zones",
                                    labels=common_labels),
            "cache_dhcp_ranges":
                CounterMetricFamily(name=f"{IBMetricDefinition.prefix}cache_dhcp_ranges",
                                    documentation=f"{IBMetricDefinition.help_prefix}number of dhcp ranges",
                                    labels=common_labels),
        }

        return metric_definition


class IBMetric(IBMetricDefinition.Labels):
    def __init__(self):
        super().__init__()
        self.cache_collect: float = 0
        self.cache_collect_failed: float = 0
        self.cache_collect_time: float = 0
        self.cache_members: float = 0
        self.cache_nodes: float = 0
        self.cache_zones: float = 0
        self.cache_dhcp_ranges: float = 0


class InfobloxMetrics(Transform):
    def __init__(self, cache: Cache):
        self.cache = cache
        self.all_metrics: List[IBMetric] = []

    def metrics(self):

        metrics_list = IBMetricDefinition.metrics_definition()
        for attribute in metrics_list.keys():
            for m in self.all_metrics:
                metrics_list[attribute].add_metric(m.get_label_values(),
                                                   m.__dict__.get(attribute))

        for m in metrics_list.values():
            yield m

    def parse(self):
        metrics: Dict[str, IBMetric] = {}

        for master, value in self.cache.get_collect_count().items():
            if master not in metrics:
                metrics[master] = IBMetric()
                metrics[master].add_label(MASTER, master)
            metrics[master].cache_collect = value

        for master, value in self.cache.get_collect_count_failed().items():
            if master not in metrics:
                metrics[master] = IBMetric()
                metrics[master].add_label(MASTER, master)
            metrics[master].cache_collect_failed = value

        for master, value in self.cache.get_collect_time().items():
            if master not in metrics:
                metrics[master] = IBMetric()
                metrics[master].add_label(MASTER, master)
            metrics[master].cache_collect_time = value

        for master, types in self.cache.get_all().items():
            if master not in metrics:
                metrics[master] = IBMetric()
                metrics[master].add_label(MASTER, master)

            for type_name, type in types.items():
                if type_name == MEMBERS:
                    metrics[master].cache_members = len(type)
                if type_name == NODES:
                    metrics[master].cache_nodes = len(type)
                if type_name == ZONES:
                    metrics[master].cache_zones = len(type)
                if type_name == DHCP_RANGES:
                    metrics[master].cache_dhcp_ranges = len(type)

        self.all_metrics.extend(list(metrics.values()))



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""

DISCOVERY_CONFIG = 'INFOBLOX_DISCOVERY_CONFIG'
DISCOVERY_PROMETHEUS_SD_FILE_DIRECTORY = 'INFOBLOX_DISCOVERY_PROMETHEUS_SD_FILE_DIRECTORY'
DISCOVERY_HOST = 'INFOBLOX_DISCOVERY_HOST'
DISCOVERY_PORT = 'INFOBLOX_DISCOVERY_PORT'
DISCOVERY_BASIC_AUTH_USERNAME = 'INFOBLOX_DISCOVERY_BASIC_AUTH_USERNAME'
DISCOVERY_BASIC_AUTH_PASSWORD = 'INFOBLOX_DISCOVERY_BASIC_AUTH_PASSWORD'
DISCOVERY_BASIC_AUTH_ENABLED = 'INFOBLOX_DISCOVERY_BASIC_AUTH_ENABLED'
DISCOVERY_LOG_FILE = 'INFOBLOX_DISCOVERY_LOG_FILE'
DISCOVERY_LOG_LEVEL = 'INFOBLOX_DISCOVERY_LOG_LEVEL'
DISCOVERY_CACHE_TTL = 'INFOBLOX_DISCOVERY_CACHE_TTL'
DISCOVERY_FETCH_INTERVAL = 'INFOBLOX_DISCOVERY_FETCH_INTERVAL'



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""

from typing import Dict, Any, Tuple
from infoblox_discovery.meta_naming import meta_label_name


class DNSServer:
    def __init__(self, host_name: str):
        self.host_name: str = host_name
        self.master: str = ''

    def _as_labels(self) -> Dict[str, str]:
        labels: Dict[str, str] = {}
        for k, v in self.__dict__.items():
            if k != "host_name":
                labels[meta_label_name(k)] = v
        return labels

    def valid(self) -> Tuple[bool, str]:
        return True, ""

    def as_prometheus_file_sd_entry(self) -> Dict[str, Any]:
        return {'targets': [f"{self.host_name}"], 'labels': self._as_labels()}


def dns_server_factory(host_name: str, master: str) -> DNSServer:
    dns_server = DNSServer(host_name)
    dns_server.master = master
    return dns_server



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""

import os
from typing import List, Any

import yaml
import logging as log
from infoblox_discovery.environments import DISCOVERY_PROMETHEUS_SD_FILE_DIRECTORY, DISCOVERY_CONFIG
from infoblox_discovery.api import InfoBlox
from infoblox_discovery.cache import MEMBERS, NODES, ZONES, DHCP_RANGES, DNS_SERVERS


def file_service_discovery():
    # Run for as file service discovery
    if not os.getenv(DISCOVERY_PROMETHEUS_SD_FILE_DIRECTORY):
        log.error(f"Env {DISCOVERY_PROMETHEUS_SD_FILE_DIRECTORY} must be set to a existing directory path")
        exit(1)
    if not os.path.exists(os.getenv(DISCOVERY_PROMETHEUS_SD_FILE_DIRECTORY)):
        log.error(f"Directory {DISCOVERY_PROMETHEUS_SD_FILE_DIRECTORY} does not exists")
        exit(1)
    with open(os.getenv(DISCOVERY_CONFIG, 'config.yml'), 'r') as config_file:
        try:
            # Converts yaml document to python object
            config = yaml.safe_load(config_file)

        except yaml.YAMLError as err:
            log.error("Parse config", extra={"error": str(err)})

    for ib in config.get('infoblox'):
        infoblox = InfoBlox(ib)
        if MEMBERS in ib.get('discovery'):
            members, nodes, dns_servers = infoblox.get_infoblox_members()
            write_sd_file(members, ib['master'], MEMBERS)
            write_sd_file(nodes, ib['master'], NODES)
            write_sd_file(dns_servers, ib['master'], DNS_SERVERS)
        if ZONES in ib.get('discovery'):
            zones = infoblox.get_infoblox_zones()
            write_sd_file(zones, ib['master'], ZONES)
        if DHCP_RANGES in ib.get('discovery'):
            dhcp_ranges = infoblox.get_infoblox_dhcp_ranges()
            write_sd_file(dhcp_ranges, ib['master'], DHCP_RANGES)


def write_sd_file(objects, prefix: str, type: str):
    prometheus_file_sd: List[Any] = []
    for key, member in objects.items():
        prometheus_file_sd.append(member.as_prometheus_file_sd_entry())

        # Generate configuration
    with open(f"{os.getenv(DISCOVERY_PROMETHEUS_SD_FILE_DIRECTORY)}/infoblox_{prefix}_{type}.yaml", 'w') as config_file:
        try:
            yaml.safe_dump(prometheus_file_sd, config_file)
        except yaml.YAMLError as err:
            log.error("Write prometheus file sd", extra={"error": str(err)})



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2023  Anders Håål

    This file is part of infoblox-discovery.

    infoblox-discovery is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    infoblox-discovery is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with infoblox-discovery.  If not, see <http://www.gnu.org/licenses/>.

"""

import abc
from typing import List

from infoblox_discovery.exceptions import DiscoveryException


class LabelsBase:
    """
    A container class to hold labels
    """
    def __init__(self):
        self.labels = {}

    def add_label(self, key: str, value: str):
        if key in self.labels.keys():
            self.labels[key] = value
        else:
            raise DiscoveryException(message=f"The label {key} is not part of defined {self.labels}", status=500)

    def get_label_keys(self) -> List[str]:
        """
        Get all label keys as a list in the specified order
        :return:
        """
        return list(self.labels.keys())

    def get_label_values(self) -> List[str]:
        """
        Get all label values as a list in the specified order
        :return:
        """
        return list(self.labels.values())


class Transform(metaclass=abc.ABCMeta):

    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, 'parse') and hasattr(subclass, 'metrics') or NotImplemented

    @abc.abstractmethod
    def parse(self):
        """
        Use this method to manage the data into a format used to
        create metrics
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def metrics(self):
        """
        Use this method to create the metrics values
        Return should be a yield like
        for m in metrics.values():
            yield m
        :return:
        """
        raise NotImplementedError
